/*
    Nonparametric Minimum Free Energy Path (NMFEP) Algorithm (Mahmoud Moradi, 1/29/2014)
    Finds the minimum free energy path between two conformations in a multidimensional free energy map.
*/

#include <stdio.h>
#include <ctype.h>
#include <cmath>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <limits>

using namespace std;

namespace {

    int T = 10000; // Maximum path length (to avoid freezing)
    double MAX = numeric_limits<double>::max();

    //General variables
    int n = 1; // number of dimensions
    int N = 1; // number of data points
    
    double* fe; // free energies
    double* X; // collective variables

    // Minimum-finding-algorithm variables
    int find_minima = 1; // locate the minima from guess
    int block = 1; // search for the minima in a block (hypercube) rather than hypersphere
    double* sigma; // the widths of coordinates (related to the metric; currently is a 1D vector)
    double* range; // used only for the block search, if specified
    int guess [2] = {0,1}; // initial guess for the minima (the indices)
    int minima [2] = {0,1}; // actual minima (the indices)

    // NMFEP algorithm variables
    int find_path = 1; // locate the path (number of times to try)
    double small = 0.5; // small and large determine the range to look for the next step (the coordinates are scaled according to sigma's)
    double large = 1.0;
    double tolerance = 0.0; // the distance tolerated to deviate from monotonic move towards the product
    
    // Output filenames
    string minimafile = "minima.txt"; // filename to report minima and their free energies
    string pathfile = "path.txt"; // filename to report the path and the free energies

    double dist(int a, int b) {
        double tot = 0;
        for (int i=0; i<n; i++)
            tot+=((X[a*n+i]-X[b*n+i])*(X[a*n+i]-X[b*n+i]))/(sigma[i]*sigma[i]);
        return sqrt(tot);
    }
    
	// reading parameters from standard input
	// '()' gives the default value
	// the parameters will be written in the standard error
    void readparm() {

        string key;
	int value;
	
        std::cerr << "Reading the parameters ...\n";
        while(!std::cin.eof()) {
            std::cin >> key;
            // PATHFILE (path.txt)
            if (key=="PATHFILE") {
                std::cin >> pathfile;
                std::cerr << "Path Filename = " << pathfile << "\n";
            // MINIMAFILE (minima.txt)
            } else if (key=="MINIMAFILE") {
                std::cin >> minimafile;
                std::cerr << "Minima Filename = " << minimafile << "\n";
             } else if (key=="NOMINIMA") {
                find_minima = 0;
                std::cerr << "Minima will not be searched.\n";
             } else if (key=="NOCUBE") {
                block = 0;
                std::cerr << "Minima will be searched with a (hyper)sphere rather than a (hyper)cube.\n";
            // NMFEP Algorithm Parameters
            // SIGMA sigma_0 sigma_1 ... (1 1 ...) 
            // STEPS small large (0.5 1)
             } else if (key=="NOPATH") {
                find_path = 0;
                std::cerr << "NMFEP will not be searched.\n";
             } else if (key=="ALLOWJUMP") {
                std::cin >> value;
                find_path = value+1;
                std::cerr << value << " time(s) jump attemp will be allowed.\n";
            } else if (key=="SIGMA") {
                for (int i=0; i<n; i++) {
                    std::cin >> sigma[i];
                    range[i]=sigma[i];
                    std::cerr << "Sigma_" << i << " = " << sigma[i] << "\n";
                }
            } else if (key=="RANGE") {
                for (int i=0; i<n; i++) {
                    std::cin >> range[i];
                    std::cerr << "range_" << i << " = " << range[i] << "\n";
                }
            } else if (key=="MINIMA") {
                std::cin >> guess[0] >> guess[1];
                std::cerr << "reactant index : " << guess[0] << "\n";
                std::cerr << "product index : " << guess[1] << "\n";
            } else if (key=="TOLERANCE") {
                std::cin >> tolerance;
                std::cerr << tolerance << " deviation from monotonic move is tolerated.\n";
            } else if (key=="PATHLENGTH") {
                std::cin >> T;
                std::cerr << "Maximum path length = " << T << "\n";
            } else if (key=="STEPS") {
                std::cin >> small >> large;
                std::cerr << "small step = " << small << "\n";
                std::cerr << "large step = " << large << "\n";
            }
        }
    }
}

// Main

int main(int argc,char *argv[]) {

    // Reading the arguments (file names)
    char* datafile = ""; // data file (data points)
    int argi = 0;
    std::cerr << "Parsing the arguments ...\n";
    while (++argi < argc) {
        std::string argstr(argv[argi]);
        if ( argstr == "--freeenergy" || argstr == "-f" ) {
            datafile = argv[++argi];
        } else if ( argstr == "--dimension" || argstr == "-d" ) {
            n = atoi(argv[++argi]);
        } else if ( argstr == "--points" || argstr == "-p" ) {
            N = atoi(argv[++argi]);
        }
    }
    
    std::cerr << "A "<< n <<"-dimensional free energy will be imported from " << datafile << " ( " << N << " data points ).\n";
    
    sigma = new double[n];
    range = new double[n];
    for (int i=0;i<n;i++) {
	sigma[i]=1.0;
	range[i]=1.0;
    }
    // Reading the parameters
    readparm();

    //
    // Importing the data
    //

    fe = new double[N];
    X = new double[N*n];

    if (datafile!="") {
	std::cerr << "Reading the free energy file ...\n";
      	ifstream fefile (datafile);
      	double x;
      	int i=0;
        int t=0;
        while (!fefile.eof()&&t<N) {
            if (i<n) {
                fefile >> X[t*n+i];
                i++;
            } else {
                fefile >> fe[t];
                i=0;
                t++;
            }
        }
        fefile.close();
        std::cerr << t << " data points read.\n";
        N=t;
    } else {
        std::cerr << "No data is available!\n";
        exit(1);
    }

    //
    // Finding the minima
    //
    for (int j=0; j<2; j++) {
        if (j)
    	    std::cerr << "product [";
    	else
    	    std::cerr << "reactant [";
        std::cerr << guess[j] << "] = ";
        for (int i=0;i<n;i++) std::cerr << X[guess[j]*n+i] << " ";
        std::cerr << "\n";
        minima[j]=guess[j];
    }

    if (find_minima) {
        std::cerr << "Finding the minima ...\n";
        for (int t=0; t<N; t++)
            for (int j=0; j<2; j++)
                if (fe[t]<fe[minima[j]]) {
            	    if (block) {
                	int i=0;
                	while ((abs(X[t*n+i]-X[guess[j]*n+i])<range[i]||range[i]<0)&&i<n) i++;
                	if (i==n) minima[j]=t;
            	    } else if (dist(t,guess[j])<1.0) minima[j]=t;
                } 
    }
    
    const char* MINIMAFILE = minimafile.c_str();
    ofstream minima_list (MINIMAFILE);
    for (int j=0; j<2; j++) {
        minima_list << j << " " << minima[j] << " " << fe[minima[j]] << " ";
        for (int i=0; i<n; i++)
            minima_list << X[minima[j]*n+i] << " ";
        minima_list << "\n";
    }
    minima_list.close();

// Finding the NMFEP between two minima

    if (find_path) {

        std::cerr << "Finding the NMFEP path!\n";

        const char* PATHFILE = pathfile.c_str();
        ofstream path (PATHFILE);
    
        double D0;
        double D1;
        double D2 = dist(minima[0],minima[1]);
        int current = -1;
        int s = 1;
        // reactant
        path << 0 << " " << minima[0] << " " << fe[minima[0]] << " ";
        for (int i=0; i<n; i++)
            path << X[minima[0]*n+i] << " ";
        path << "\n";
	// path
        while (s<T && D2>large) {
	    int attempt=1;
	    while(attempt<=find_path&&current<0) {
        	for (int t=0; t<N; t++)
            	    if (fe[t]<fe[current]||current<0) {
                	D0=dist(t,minima[0]);
                	if (D0>small&&D0<attempt*large) {
                	    D1=dist(t,minima[1]);
                	    if (D1-D2<tolerance) current=t;
            		}
            	    }
        	if (current<0) attempt++;
    	    }
    	    if (current<0) {
                std::cerr << "Step "<< s <<": No data point in the range was found.\n";
                exit(1);
    	    } else {
                if (attempt>1) std::cerr << "Step " << s << " required " << attempt-1 << " jump(s).\n";
                path << s << " " << current << " " << fe[current] << " ";
                for (int i=0; i<n; i++)
            	    path << X[current*n+i] << " ";
            	path << "\n";
            	D2=dist(current,minima[1]);
            	minima[0] = current;
            	current = -1;
            	s++;
    	    }
        }
        // product
        path << s << " " << minima[1] << " " << fe[minima[1]] << " ";
        for (int i=0; i<n; i++)
            path << X[minima[1]*n+i] << " ";
        path << "\n";
            
        path.close();
    }
    
    delete[] fe;
    delete[] X;

    std::cerr << "Done!\n";

}
